﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=StaticCode
Version=8.8
@EndOfDesignText@
'Static code module
'CUBIC SPLINE For EQUIDISTANT POINTS

'https://www.bragitoff.com/2018/02/cubic-spline-piecewise-interpolation-c-program/
' Function that performs Gauss-Elimination And returns the Upper triangular matrix And solution of equations:
'1. Pass the augmented matrix (a) As the parameter, And calculate And store the upperTriangular(Gauss-Eliminated Matrix) in it.


'============================================================================================
Sub Process_Globals
	Type splineCoeffs_t(n As Int, a As Double, b As Double, c As Double, d As Double)
End Sub


'============================================================================================
private Sub gaussEliminationLS(m As Int, n As Int, a(,) As Double, x() As Double)
	Dim i, j, k  As Int
	Dim temp As Double
	Dim term As Double
	
	For i = 0 To m - 2 'Check the 2
		'Partial Pivoting
		For k = i + 1 To m - 1
			'If diagonal element(absolute vallue) is smaller than any of the terms below it
			If Abs(a(i,i)) < Abs(a(k,i)) Then
				'Swap the rows
				For j = 0 To n - 1
					temp = a(i,j)
					a(i,j) = a(k,j)
					a(k,j) = temp
				Next
			End If
		Next
		'Begin Gauss Elimination
		For k = i + 1 To m - 1
			term = a(k,i) / a(i,i)
			For j = 0 To (n -1)
				a(k,j) = a(k,j) - term * a(i,j)
			Next
		Next
	Next
	
	'Begin Back-substitution
	For i = (m - 2) To  0 Step -1
		x(i) = a(i, n-1)
		For j = i + 1 To n - 2
			x(i) = x(i) - a(i,j) * x(j)
		Next
		x(i) = x(i) / a (i,i)
	Next
End Sub


'============================================================================================
'Cubic Spline coefficients calculator

Private Sub cSCoeffCalc(n As Int, h As Double, sig() As Double, y() As Double, _
						a() As Double, b()As Double, c() As Double,	d() As Double)
	Dim i As Int
	
	For i = 0 To n - 1
		d(i) = y(i)
		b(i) = sig(i) / 2.0
		a(i) = (sig(i + 1) - sig(i)) / (h * 6)
		c(i) = (y(i + 1) - y(i)) / h - h * (2 * sig(i) + sig(i + 1)) / 6.0
	Next
End Sub


'============================================================================================
'Function To generate the tridiagonal augmented matrix 
'For cubic spline For equidistant data-points

private Sub tridiagonalCubicSplineGen(n As Int, h As Double, a(,) As Double, y() As Double)
	Dim i As Int
	
	For i = 0 To n - 2
		a(i,i) = 4
	Next

	For i = 0 To n - 3
		a(i, i+1) = 1
		a(i+1, i) = 1
	Next
	
	For i = 0 To n - 2
'		Log("i = " & i)
		a(i, n-1) = (y(i+2) - 2 * y(i+1) + y(i)) * 6 / h / h
	Next
End Sub


'============================================================================================
'Function that copies the elements of a matrix To another matrix
'Parameters: rows(m),columns(n),matrix1[m][n] , matrix2[m][n]

Private Sub copyMatrix(m As Int, n As Int, matrix1(,) As Double, matrix2(,) As Double)
	Dim i,j As Int
	
	For i = 0 To m - 1
		For j = 0 To n - 1
			matrix2(i, j) = matrix1(i, j)
		Next
	Next
End Sub


'============================================================================================
Public Sub calcSpline(x() As Double, y() As Double) As List
	Dim splineList As List
	splineList.Initialize
	
	Dim h As Double = 1.0
	Dim n As Int  = x.Length - 1
	Dim a(n) As Double
	Dim b(n) As Double
	Dim c(n) As Double
	Dim d(n) As Double
	Dim sig(n+1) As Double
	Dim sigTemp(n-1) As Double
	sig(0) = 0
	sig(n) = 0
	
	Dim tri(n - 1, n) As Double
	
	tridiagonalCubicSplineGen(n, h, tri, y)
	
	'Perform Gauss Elimination
	gaussEliminationLS(n-1, n, tri, sigTemp)
	
	For i = 1 To n-1
		sig(i) = sigTemp(i-1)
	Next
	
	cSCoeffCalc(n, h, sig, y, a, b, c, d)
	
	For i = 0 To n - 1
		Dim s As splineCoeffs_t
		s.Initialize
		s.n = i
		s.a = a(i)
		s.b = b(i)
		s.c = c(i)
		s.d = d(i)
		splineList.Add(s)
	Next
	Return splineList
End Sub


'============================================================================================
'Interpolate vector y to vector yInterp using the nRatio > 1
Public Sub interpCubicSplineEqX(y() As Double, yInterp() As Double, nRatio As Int) As Boolean

	If y.Length < 4 Or nRatio < 2 Or yInterp.Length < nRatio * y.Length Then
		Return False
	End If
	
	Private splineList As List
	splineList.Initialize
	
	Dim n As Int = y.Length 
	
	'Create dummy x values for the vector remembering they are quidistant
	'Assume they are unity. The user can decide outside of this what x basis they want.
	Private xPoints(n) As Double
	
	For i = 0 To n-1
		xPoints(i) = i
	Next
	
	'Calc spline coefficients
	splineList = calcSpline(xPoints, y)

	'Interpolate y into yInterp using the interpolation ration nRatio
	Private xt As Double
	For i = 0 To  n - 2
		Dim s As splineCoeffs_t
		s.Initialize
		
		s = splineList.Get(i)
		
		For k = 0 To nRatio - 1
			xt = k / nRatio
			If k = 0 Then
				yInterp(nRatio * i) = s.d
			Else
				yInterp(nRatio * i + k) = s.a * xt * xt * xt + s.b * xt * xt + s.c * xt + s.d
			End If
		Next
	Next
	yInterp(nRatio * i) = y(n - 1)
	Return True
End Sub


'============================================================================================